/*
 * Copyright (C) 2010  Peter Bailis <pbailis@eecs.harvard.edu>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <assert.h>
#include <float.h>
#include <cmath>
#include <iostream>
#include "Bee.h"
#include "RunSettings.h"

#define DEG_TO_RADIANS(x) ((float)((x) * 0.0174532925))
#define MAX(a, b) (a > b ? a : b)

using namespace std;

Bee::Bee(const unsigned id, Position *initPos, beeType role, Colony *beeColony, unsigned inputWorldSize)
{
    myPosition = initPos;
    perceivedPosition = new Position(initPos->getXCoord(), initPos->getYCoord());

    thinkWeReachedDestination = false;

    forgetFlowerThreshold = 1;
    danceFlowerThreshold = 4;
    desiredPosition = new Position(initPos->getXCoord(), initPos->getYCoord());
    flightEnergyUsed = 0;
    flightEnergyLeft = beeColony->getEnergy(0);
    knownFlowerPosition = new Position(initPos->getXCoord(),initPos->getYCoord());
    realKnownFlowerPosition = new Position(initPos->getXCoord(),initPos->getYCoord());
    hasKnownFlower = false;
    carryingNectar = false;
    arrivedAtFlowerLocation = false;
    thinkAtHome = false;
    newRandomPosition = false;
    myRole = role;
    myColony = beeColony;
    myId = id;
    worldSize = inputWorldSize;
    returnHome = false;
    dancing = false;
    doDanceNow = false;
    randomWalking = false;
    runAccounted = true;
    danceTimeLeft = 0;
    knownFlowerQuality = 0;
    nectarQuantity = 0;
}

Bee::~Bee()
{
    delete myPosition;
    delete desiredPosition;
    delete perceivedPosition;
    delete knownFlowerPosition;
    delete realKnownFlowerPosition;
}

const unsigned Bee::getId()
{
    return myId;
}

const Position *Bee::getPosition()
{
    return myPosition;
}

const Position *Bee::getRealKnownFlowerPosition()
{
    if(!hasKnownFlower)
    {
        return NULL;
    }

    return realKnownFlowerPosition;
}

const beeType Bee::getRole()
{
    return myRole;
}

const bool Bee::hasNectar()
{
    return carryingNectar;
}

void Bee::updateBee(FlowerLocator *flowerLocations)
{
    //if we're at the colony
    if (*myPosition == *(myColony->getPosition()))
    {

      //this is just for stat logging--does not functionally change update()
        if(myRole == FORAGER && !runAccounted)
        {
            runAccounted = true;
            totalForageTrips++;

            if(!carryingNectar)
            {
              unsuccessfulReturns++;
              hasKnownFlower = false;
            }
        }
	//more logging stuff
        if(myRole == FORAGER && carryingNectar)
        {
            if(nectarQuantity == 1)
            {
                successfulReturnLo++;
            }
            else
            {
                successfulReturnHigh++;
            }
        }
        else if(myRole == SCOUT)
        {
            myColony->reportScout();
        }


        //get new energy, report the amount we've used to the colony
        flightEnergyLeft = myColony->getEnergy(flightEnergyUsed);
        flightEnergyUsed = 0;
        returnHome = false;
	
        if (carryingNectar)
        {
	    myColony->deliverFood(nectarQuantity);
	    carryingNectar = false;
	    doDanceNow = true;
        }

	if(myRole == SCOUT && hasKnownFlower)
	{
	  doDanceNow = true;
	}

        myRole = FORAGER;
        //reset our status variables
        randomWalking = false;
        thinkAtHome = false;
        arrivedAtFlowerLocation = false;

        //correct our position
        perceivedPosition->setXCoord(myPosition->getXCoord());
        perceivedPosition->setYCoord(myPosition->getYCoord());
        //waggle dancing--keep dancing
        if(dancing)
        {
            assert(runSettings->doSharePositionalInformation() != NODANCE);

            danceTimeLeft--;

            if(!danceTimeLeft)
            {
                dancing = false;
                myColony->endDance(knownFlowerPosition);
            }
            else
            {
                return;
            }
        }
        else if(doDanceNow)
        {
	    doDanceNow = false;
	    assert(forgetFlowerThreshold <= danceFlowerThreshold);

            float flowerValue = knownFlowerQuality;

            //so bad we won't remember the flower
            if(flowerValue <= forgetFlowerThreshold)
            {
                hasKnownFlower = false;

                float change = (float)(forgetFlowerThreshold-flowerValue)/2;

                if(forgetFlowerThreshold != 0)
                {
                    if(change > forgetFlowerThreshold)
                        forgetFlowerThreshold = 0;
                    else
                        forgetFlowerThreshold -= change;
                }
            }
            //we'll follow it, but we won't advertise it
            else if(flowerValue <= danceFlowerThreshold)
            {
                float change = max((float).1, (float)(danceFlowerThreshold-flowerValue)/2);

                if(danceFlowerThreshold != 0)
                {
                    if(change > danceFlowerThreshold)
                        danceFlowerThreshold = 0;
                    else
                        danceFlowerThreshold -= change;

                }

                float fchange = max((float) .1, (float)(flowerValue-forgetFlowerThreshold)/2);

		if(forgetFlowerThreshold > danceFlowerThreshold)
		{
		  forgetFlowerThreshold = danceFlowerThreshold;
		}
                else if(forgetFlowerThreshold + fchange < danceFlowerThreshold)
                {
		  forgetFlowerThreshold += fchange;
                }
		else
		{
		  forgetFlowerThreshold = danceFlowerThreshold;
		}
            }
            //dance
            else
            {
                danceTimeLeft = flowerValue;

                float change = (float)(flowerValue-danceFlowerThreshold)/2;

                danceFlowerThreshold += change;

                if(danceTimeLeft && runSettings->doSharePositionalInformation() != NODANCE)
                {
                    myColony->startDance(knownFlowerPosition);
                    dancing = true;
                    return;
                }
            }
        }

        if (myRole == FORAGER)
        {
            //get a flower if we don't have one
            if (!hasKnownFlower)
            {
                if(runSettings->doSharePositionalInformation() == NODANCE)
                {
                    myRole = SCOUT;
		    hasKnownFlower = false;
                    //make scouts more likely to dance
                    forgetFlowerThreshold /= 2;
                    danceFlowerThreshold /= 2;
                }
                else if((float)random()/RAND_MAX < runSettings->getUnemployedScoutProbability())
                {
                    myRole = SCOUT;
		    hasKnownFlower = false;
                    desiredPosition->setXCoord(myColony->getPosition()->getXCoord());
                    desiredPosition->setYCoord(myColony->getPosition()->getYCoord());

                    //make scouts more likely to dance
                    forgetFlowerThreshold /= 2;
                    danceFlowerThreshold /= 2;
                }
                else
                {
                    Position *getFlowerPosition = myColony->getFlowerAssignment();

                    //update our known flower position
                    if(getFlowerPosition)
                    {
                        unsigned setting = runSettings->doSharePositionalInformation();

                        //if we're actually following dance flower positions, or if we don't have a flower
                        if(setting == DANCE_INFO)
                        {
                            knownFlowerPosition->setXCoord(getFlowerPosition->getXCoord());
                            knownFlowerPosition->setYCoord(getFlowerPosition->getYCoord());
                            hasKnownFlower = true;
                        }
                        //go back to previously known flower
                        else if(setting == DANCE_INTERNALINFO)
                        {
                            if(!(*knownFlowerPosition == *(myColony->getPosition())))
                                hasKnownFlower = true;
                            else
                            {
                                myRole = SCOUT;
				hasKnownFlower = false;
                                desiredPosition->setXCoord(myColony->getPosition()->getXCoord());
                                desiredPosition->setYCoord(myColony->getPosition()->getYCoord());
				//make scouts more likely to dance
				forgetFlowerThreshold /= 2;
				danceFlowerThreshold /= 2;
                            }
                        }
                        else if(setting == DANCE_SCOUT)
                        {
                                myRole = SCOUT;
				hasKnownFlower = false;
                                desiredPosition->setXCoord(myColony->getPosition()->getXCoord());
                                desiredPosition->setYCoord(myColony->getPosition()->getYCoord());
				//make scouts more likely to dance
				forgetFlowerThreshold /= 2;
				danceFlowerThreshold /= 2;
                        }
                        else if(setting == NODANCE)
                        {
                            assert(false);
                        }
                        else
                        {
                            assert(false);
                        }

                    }
                    //be dormant
                    else
                    {
                        return;
                    }
                }
            }
        }
    }

    if (myRole == FORAGER)
    {
        bool atKnownFlowerPosition = (hasKnownFlower && *perceivedPosition == (*knownFlowerPosition));

        if (!carryingNectar && (atKnownFlowerPosition ||
             arrivedAtFlowerLocation))
        {
            arrivedAtFlowerLocation = true;

            //try to find a flower and harvest it
            if(!findAndHarvestFlower())
            {
                if(atKnownFlowerPosition)
                {
                    if(hasKnownFlower && !carryingNectar)
                    {
                        unsuccessfulHeadings++;
                    }
                    knownFlowerQuality = 0;
                    hasKnownFlower = false;
                }

                //try to detect another one
                detectFlower();

                if(!hasKnownFlower && flightEnergyLeft > 0)
                {
                    flightEnergyLeft--;
                    flightEnergyUsed++;
                    //if we didn't find it, levy walk until we do, or until we run out of energy
                    if(!randomWalking)
                    {
                        desiredPosition->setXCoord(perceivedPosition->getXCoord());
                        desiredPosition->setYCoord(perceivedPosition->getYCoord());
                    }

                    randomWalk();
                    return;
                }
                else
                {
                    arrivedAtFlowerLocation = false;
                }
            }
        }
    }
    else if (myRole == SCOUT)
    {
        this->detectFlower();
    }

    this->moveBee();

    return;
}

void Bee::moveBee()
{
    flightEnergyLeft--;
    flightEnergyUsed++;

    if (flightEnergyLeft <= 0)
    {
        goHome();
        return;
    }

    if (myRole == FORAGER)
    {
      runAccounted = false;

        if (carryingNectar)
        {
            goHome();
        }
        else if(hasKnownFlower)
        {
            returnHome = false;
            this->hone(knownFlowerPosition);
        }
    }
    else if (myRole == SCOUT)
    {
        this->scoutMove();
    }
}

const unsigned Bee::getForgetFlowerThreshold() const
{
    return forgetFlowerThreshold;
}

const unsigned Bee::getDanceFlowerThreshold() const
{
    return danceFlowerThreshold;
}

static float perturbValue(float value, float noise)
{
    //want to get between [value-noise, value+noise]
//    cout << (1-(((float)rand()/RAND_MAX)*noise*2 - noise));
    return value*(1-(((float)rand()/RAND_MAX)*noise*2 - noise));
}

static float perturbSensing(float value)
{
    return perturbValue(value, runSettings->getSensingNoise());
}

static float perturbActuation(float value)
{
    return perturbValue(value, runSettings->getActuationNoise());
}

void Bee::hone(const Position *desired)
{
    float perceivedDistance, perceivedAngle,
          desiredDistance = runSettings->getBeeFlightSpeed();

    perceivedAngle = perceivedPosition->getAngle(*desired);
    perceivedDistance = perceivedPosition->getDistance(*desired);

    if(desiredDistance > perceivedDistance)
    {
        desiredDistance = perceivedDistance;
    }

    //calculate real movement, with actuation error
    float actualAngle = perturbActuation(perceivedAngle);
    float actualDistance = perturbActuation(desiredDistance);

    myPosition->incrementXCoord(cos(actualAngle)*actualDistance);
    myPosition->incrementYCoord(sin(actualAngle)*actualDistance);

    //calculate perceived movement, with sensing error

    perceivedAngle = perturbSensing(actualAngle);
    perceivedDistance = perturbSensing(actualDistance);

    perceivedPosition->incrementXCoord(cos(perceivedAngle)*desiredDistance);
    perceivedPosition->incrementYCoord(sin(perceivedAngle)*desiredDistance);

    float myXCoord = myPosition->getXCoord(), myYCoord = myPosition->getYCoord();

    if(myXCoord < 0)
        myPosition->setXCoord(0);
    else if(myXCoord > worldSize)
        myPosition->setXCoord(worldSize);

    if(myYCoord < 0)
        myPosition->setYCoord(0);
    else if(myYCoord > worldSize)
        myPosition->setYCoord(worldSize);


}

void Bee::goHome(void)
{
    if(!randomWalking)
    {
        desiredPosition->setXCoord(myColony->getPosition()->getXCoord());
        desiredPosition->setYCoord(myColony->getPosition()->getYCoord());
    }

    if(myRole == FORAGER && !carryingNectar)
    {
        findAndHarvestFlower();
    }

    if(myColony->getPosition()->getDistance(*myPosition) < runSettings->getBeeColonySightRadius())
    {
        perceivedPosition->setXCoord(myPosition->getXCoord());
        perceivedPosition->setYCoord(myPosition->getYCoord());
        randomWalking = false;
        this->hone(myColony->getPosition());

        return;
    }
    if(!(*perceivedPosition == *(myColony->getPosition())) && !thinkAtHome)
    {
        this->hone(myColony->getPosition());
    }
    else
    {
        thinkAtHome = true;
        randomWalk();
    }
}

void Bee::scoutMove()
{
    float myXCoord = myPosition->getXCoord();
    float myYCoord = myPosition->getYCoord();

    if((myXCoord == myColony->getPosition()->getXCoord()) && (myYCoord == myColony->getPosition()->getYCoord()))
    {
        desiredPosition->setXCoord(perceivedPosition->getXCoord());
        desiredPosition->setYCoord(perceivedPosition->getYCoord());
    }

    if(returnHome || myXCoord > worldSize || myYCoord > worldSize || myXCoord < 0 || myYCoord < 0)
    {
        returnHome = true;
        goHome();
        return;
    }

    scoutStrategy chosenStrategy = runSettings->getChosenScoutStrategy();

    if(chosenStrategy == RANDOMWALK)
    {
        randomWalk();
    }
    else if(chosenStrategy == SPIRAL)
    {
        if(*myPosition == *(myColony->getPosition()))
        {
            bearing = random() % 360;
            velocity = runSettings->getBeeFlightSpeed();
            desiredPosition->setXCoord(myXCoord);
            desiredPosition->setYCoord(myYCoord);
            getNewSpiralDestination();
        }
        else if(*perceivedPosition == *desiredPosition)
        {
            bearing = (bearing + runSettings->getSpiralAngularVelocityIncrease()) % 360;
            velocity += runSettings->getSpiralVelocityIncrease();
            getNewSpiralDestination();
        }

        hone(desiredPosition);
    }

    return;
}

float Bee::getFlowerValue(const Position *position, float flowerQuality, float quantity)
{
    if(quantity < 0)
        return 0;

    float ret = ((maxDistance-position->getDistance(*(myColony->getPosition())))/
		 maxDistance)*flowerQuality*runSettings->getBeeDanceLengthMultiplier();

    //perceived distance could be greater than max distance; if this is the case,
    //just return 0
    if(ret < 0)
      ret = 0;

    return ret;
}

void Bee::randomWalk(void)
{
    float myXCoord = myPosition->getXCoord();
    float myYCoord = myPosition->getYCoord();

    randomWalking = true;

    if(!(*desiredPosition == *perceivedPosition) &&
       !(myXCoord > worldSize || myYCoord > worldSize || myXCoord < 0 || myYCoord < 0))
    {
        newRandomPosition = false;
        hone(desiredPosition);
        return;
    }

    newRandomPosition = false;

    float flatRandom = ((float)random())/RAND_MAX;
    float angle = random() % 360;

    float maxDistance = worldSize/2;
    float minDistance = runSettings->getBeeFlightSpeed();
    float degree = -2;

    //http://stackoverflow.com/questions/918736/random-number-generator-that-produces-a-power-law-distribution
    //http://mathworld.wolfram.com/RandomNumber.html
    float newMagnitude = pow((pow(maxDistance, degree+1) - pow(minDistance, degree+1))*flatRandom + 1, (1/(degree+1)));

    float newXPosition = perceivedPosition->getXCoord() + sin(DEG_TO_RADIANS(angle))*newMagnitude;
    float newYPosition = perceivedPosition->getYCoord() + cos(DEG_TO_RADIANS(angle))*newMagnitude;

    if(newXPosition < 0)
        newXPosition = 0;
    else if(newXPosition > worldSize)
        newXPosition = worldSize;
    if(newYPosition < 0)
        newYPosition = 0;
    else if(newYPosition > worldSize)
        newYPosition = worldSize;

    desiredPosition->setXCoord(newXPosition);
    desiredPosition->setYCoord(newYPosition);

    hone(desiredPosition);

    return;
}

void Bee::getNewSpiralDestination(void)
{
    //bearing measured such that vertical is 0 degrees
    float x = sin(DEG_TO_RADIANS(bearing))*velocity;
    float y = cos(DEG_TO_RADIANS(bearing))*velocity;

    desiredPosition->incrementXCoord(x);
    desiredPosition->incrementYCoord(y);
}

void Bee::detectFlower()
{
    vector<Flower *> *toConsider = flowerLocations->findFlowersByPosition(myPosition, runSettings->getBeeVisionRadius());
    vector<Flower *>::iterator flowerIterator;
    Flower *newFlower = NULL;

    float myXCoord = myPosition->getXCoord();
    float myYCoord = myPosition->getYCoord();
    float perceivedXCoord = perceivedPosition->getXCoord();
    float perceivedYCoord = perceivedPosition->getYCoord();
    float xOffset, yOffset;
    float thisValue;

    if(toConsider)
    {
        for(flowerIterator = toConsider->begin();
            flowerIterator != toConsider->end();
            flowerIterator++)

            newFlower = *flowerIterator;

            thisValue = this->getFlowerValue(newFlower->getPosition(), newFlower->getQuality(), newFlower->getQuantity());

            if (!hasKnownFlower || knownFlowerQuality < thisValue)
            {
                xOffset = newFlower->getPosition()->getXCoord()-myXCoord;
                yOffset = newFlower->getPosition()->getYCoord()-myYCoord;
                knownFlowerPosition->setXCoord(perceivedXCoord+xOffset);
                knownFlowerPosition->setYCoord(perceivedYCoord+yOffset);
                realKnownFlowerPosition->setXCoord(newFlower->getPosition()->getXCoord());
                realKnownFlowerPosition->setYCoord(newFlower->getPosition()->getYCoord());
                knownFlowerQuality = thisValue;
                hasKnownFlower = true;
            }
    }

    delete toConsider;
}

bool Bee::findAndHarvestFlower()
{
    vector<Flower *> *toConsider = flowerLocations->findFlowersByPosition(myPosition, 1);
    vector<Flower *>::iterator flowerIterator;
    Flower *bestFlower = NULL;
    Flower *newFlower = NULL;

    assert(!carryingNectar);

    bool ret = false;
    float newValue, bestValue = -1;

    float myXCoord = myPosition->getXCoord();
    float myYCoord = myPosition->getYCoord();
    float perceivedXCoord = perceivedPosition->getXCoord();
    float perceivedYCoord = perceivedPosition->getYCoord();
    float xOffset, yOffset;

    if(toConsider)
    {
        for(flowerIterator = toConsider->begin();
            flowerIterator != toConsider->end();
            flowerIterator++)

            newFlower = *flowerIterator;
            newValue = this->getFlowerValue(newFlower->getPosition(), newFlower->getQuality(), newFlower->getQuantity());

            if(newValue > bestValue)
            {
                bestFlower = newFlower;
                bestValue = newValue;
            }
    }

    if(bestFlower)
    {
        if(bestFlower->harvestNectar())
        {
            hasKnownFlower = true;
            carryingNectar = true;
            knownFlowerQuality = bestValue;
            nectarQuantity = newFlower->getQuality();

            xOffset = newFlower->getPosition()->getXCoord()-myXCoord;
            yOffset = newFlower->getPosition()->getYCoord()-myYCoord;
            knownFlowerPosition->setXCoord(perceivedXCoord+xOffset);
            knownFlowerPosition->setYCoord(perceivedYCoord+yOffset);
            realKnownFlowerPosition->setXCoord(newFlower->getPosition()->getXCoord());
            realKnownFlowerPosition->setYCoord(newFlower->getPosition()->getYCoord());

            //if we're the last ones
            if(bestFlower->isDead())
            {
                hasKnownFlower = false;
            }
        }

        ret = true;
    }

    delete toConsider;

    return ret;
}
