/*
 * Copyright (C) 2010  Peter Bailis <pbailis@eecs.harvard.edu>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "ConfigReader.h"
#include "Default.h"

#include <string.h>

#include <iostream>
#include <assert.h>

using namespace std;

unsigned worldSize;
float maxDistance;

ConfigReader::ConfigReader(char *inputFile)
{
    timeSteps = DEFAULT_TIME;
    nectarPerHarvest = DEFAULT_NECTAR_PER_HARVEST;
    strcpy(logFile, DEFAULT_LOGFILE);
    locationAccuracy = DEFAULT_LOCATION_ACCURACY;
    beeFlightSpeed = DEFAULT_BEE_FLIGHT_SPEED;
    energyToReplenish = DEFAULT_ENERGY_TO_REPLENISH;
    spiralVelocityIncrease = DEFAULT_SPIRAL_VELOCITY_INCREASE;
    spiralAngularVelocityIncrease = DEFAULT_SPIRAL_ANGULAR_VELOCITY_CHANGE;
    actuationNoise = DEFAULT_BEE_ACTUATION_NOISE;
    sensingNoise = DEFAULT_BEE_SENSING_NOISE;
    beeColonySightRadius = DEFAULT_BEE_COLONY_SIGHT_RADIUS;
    unemployedScoutProbability = DEFAULT_UNEMPLOYED_SCOUT_PROBABILITY;
    beeDanceLengthMultiplier = DEFAULT_BEE_DANCE_LENGTH_MULTIPLIER;
    sharePositionalInformation = DEFAULT_SHARE_POSITIONAL_INFORMATION;
    numFollowersPerDancer = DEFAULT_NUM_FOLLOWERS_PER_DANCER_PER_TIMESTEP;

    configFile.open(inputFile);

    readConfig();
}

ConfigReader::~ConfigReader()
{
    configFile.close();
    worldFile.close();
}

void ConfigReader::readConfig()
{
    char buf[100];
    char *param, *value;

    while(configFile.getline(buf, 99))
    {
        if(buf[0] == '#' || buf[0] == '\0')
            continue;

        param = strtok(buf, " ");
        value = strtok(NULL, " ");

        if(!strcmp(param, "WORLD"))
        {
            if(strcmp(value, "none"))
                worldFile.open(value);
        }
        else if(!strcmp(param, "SCOUTSTRATEGY"))
        {
            if(!strcmp(value, "SPIRAL"))
            {
                chosenStrategy = SPIRAL;
            }
            else if(!strcmp(value, "RANDOMWALK"))
            {
                chosenStrategy = RANDOMWALK;
            }
        }
        else if(!strcmp(param, "TIME"))
        {
            timeSteps = atoi(value);
        }
        else if(!strcmp(param, "LOGFILE"))
        {
            if(strcmp(value, "none"))
            {
                strcpy(logFile, value);
            }
        }
        else if(!strcmp(param, "NECTARPERHARVEST"))
        {
            nectarPerHarvest = atoi(value);
        }
        else if(!strcmp(param, "BEEFLIGHTSPEED"))
        {
            beeFlightSpeed = atoi(value);
        }
        else if(!strcmp(param, "LOCATIONACCURACY"))
        {
            locationAccuracy = atof(value);
        }
        else if(!strcmp(param, "RANDSEED"))
        {
            randSeed = atoi(value);
        }
        else if(!strcmp(param, "HIVEREPLENISHENERGY"))
        {
            energyToReplenish = atoi(value);
        }
        else if(!strcmp(param, "BEEVISIONRADIUS"))
        {
            beeVisionRadius = atof(value);
        }
        else if(!strcmp(param, "SPIRALVELOCITYINCREASE"))
        {
            spiralVelocityIncrease = atof(value);
        }
        else if(!strcmp(param, "SPIRALANGULARVELOCITYINCREASE"))
        {
            spiralAngularVelocityIncrease = atoi(value);
        }
        else if(!strcmp(param, "BEESENSINGNOISE"))
        {
            sensingNoise = atof(value);
        }
        else if(!strcmp(param, "BEEACTUATIONNOISE"))
        {
            actuationNoise = atof(value);
        }
        else if(!strcmp(param, "COLONYVISIONRADIUS"))
        {
            beeColonySightRadius = atof(value);
        }
        else if(!strcmp(param, "UNEMPLOYEDSCOUTPROBABILITY"))
        {
            unemployedScoutProbability = atof(value);
        }
        else if(!strcmp(param, "BEEDANCELENGTHMULTIPLIER"))
        {
            beeDanceLengthMultiplier = atoi(value);
        }
        else if(!strcmp(param, "FOLLOWERSPERDANCERPERTIMESTEP"))
        {
            numFollowersPerDancer = atoi(value);
        }
        else if(!strcmp(param, "SHAREPOSITIONALINFORMATION"))
        {
            sharePositionalInformation = atoi(value);
        }
        else
        {
            cout << "unknown parameter in config file: " << param << endl;
        }
    }
}

Settings *ConfigReader::getSettings(void)
{
    return new Settings(timeSteps,
                        nectarPerHarvest,
                        beeFlightSpeed,
                        energyToReplenish,
                        locationAccuracy,
                        randSeed,
                        chosenStrategy,
                        spiralVelocityIncrease,
                        beeVisionRadius,
                        actuationNoise,
                        sensingNoise,
                        beeColonySightRadius,
                        unemployedScoutProbability,
                        spiralAngularVelocityIncrease,
                        beeDanceLengthMultiplier,
                        sharePositionalInformation,
                        numFollowersPerDancer,
                        logFile);
}

World *ConfigReader::getWorld(void)
{
    World *newWorld;
    Colony *newColony;

    map<unsigned, Flower *> *flowersToAdd = new map<unsigned, Flower *>();


    char buf[100];
    char *param;

    worldFile.getline(buf,99);
    worldFile.getline(buf,99);

    worldSize = atoi(buf);

    //length of side/2 * sqrt(2), with a small bit of fudge
    maxDistance = (worldSize/2)*1.4143;

    newWorld = new World(worldSize);

    worldFile.getline(buf,99);
    worldFile.getline(buf,99);

    assert(buf[0] == 'C');

    strtok(buf, " ");
    newColony = new Colony(new Position(atoi(strtok(NULL, " ")), atoi(strtok(NULL, " "))));

    newWorld->addColony(newColony);

    while(worldFile.getline(buf, 99))
    {
        if(buf[0] == '#')
            continue;

        param = strtok(buf, " ");

        if(!strcmp(param, "B"))
        {
            unsigned id = atoi(strtok(NULL, " "));
            float x = atof(strtok(NULL, " "));
            float y = atof(strtok(NULL, " "));
            char *type = strtok(NULL, " ");
            beeType actualType = SCOUT;

            if(!strcmp(type, "SCOUT"))
            {
                actualType = SCOUT;
            }
            else if(!strcmp(type, "FORAGER"))
            {
                actualType = FORAGER;
            }

            newWorld->addBee(new Bee(id, new Position(x,y), actualType, newColony, worldSize));
        }
        else if(!strcmp(param, "F"))
        {
            unsigned id = atoi(strtok(NULL, " "));
            float x = atof(strtok(NULL, " "));
            float y = atof(strtok(NULL, " "));
            unsigned life = atoi(strtok(NULL, " "));
            unsigned quantity = atoi(strtok(NULL, " "));
            unsigned quality = atoi(strtok(NULL, " "));

            newWorld->addFlower(new Flower(id, new Position(x,y), life, quantity, quality));
        }
        else if(!strcmp(param, "NF"))
        {
            unsigned id = atoi(strtok(NULL, " "));
            float x = atof(strtok(NULL, " "));
            float y = atof(strtok(NULL, " "));
            unsigned life = atoi(strtok(NULL, " "));
            unsigned quantity = atoi(strtok(NULL, " "));
            unsigned quality = atoi(strtok(NULL, " "));
            unsigned time = atoi(strtok(NULL, " "));

            (*flowersToAdd)[time] = (new Flower(id, new Position(x,y), life, quantity, quality));
        }
        else
        {
            cout << "unknown parameter in world file: " << param << endl;
        }
    }

    newWorld->addFutureFlowers(flowersToAdd);

    return newWorld;
}
