/*
 * Copyright (C) 2010  Peter Bailis <pbailis@eecs.harvard.edu>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <stdlib.h>
#include "Default.h"
#include "World.h"

#include <algorithm>
#include <assert.h>
#include <iostream>

using namespace std;

FlowerLocator *flowerLocations;

World::World(unsigned size)
{
    mySize = size;

    //added in addColony
    colony = NULL;

    flowerLocations = new FlowerLocator();
    time = 0;
    nextGuid = 0;

    if(doLog)
    {
        debugLogger = new Logger();
        debugLogger->logWorldStats(size);
    }

    //addNBees();
    //addNFlowers();
}

World::~World()
{
    if(doLog)
        debugLogger->logFinalStats(this->totalNectarCollected(),
                                   this->totalEnergySpent());

    vector<Bee *>::iterator beeIterator;
    vector<Flower *>::iterator flowerIterator;

    for(beeIterator = bees.begin();
        beeIterator != bees.end();
        beeIterator++)
    {
        delete *beeIterator;
    }

    for(flowerIterator= flowers.begin();
        flowerIterator != flowers.end();
        flowerIterator++)
    {
        delete *flowerIterator;
    }

    delete colony;
    delete flowerLocations;
    delete flowersToAdd;

    if(doLog)
        delete debugLogger;
}

const unsigned World::getNewGuid()
{
    return nextGuid++;
}

void World::addFutureFlowers(std::map<unsigned, Flower *> *inputFlowers)
{
    flowersToAdd = inputFlowers;
}


void World::updateWorld(void)
{
    vector<Bee *>::iterator beeIterator;
    vector<Flower *>::iterator flowerIterator, deleteFlowerIterator;

    time++;

    if(doLogKnownFlowers)
        cout << time << endl;

    colony->initializeDancers();

    //randomize bees--helps with ensuring fairness of dancing
    if(!(time % 10))
    {
        random_shuffle(bees.begin(), bees.end());
    }

    if(doLog)
        debugLogger->logTimeStep(time);

    Flower *newFlower = (*flowersToAdd)[time];

    if(newFlower)
    {
        flowers.push_back(newFlower);
        flowerLocations->insertFlower(newFlower);
        if(doLog)
            debugLogger->logFlower(newFlower);
    }

    unsigned threshLow = 0;
    unsigned threshHi = 0;
    unsigned zerolo = 0;
    unsigned zerohi = 0;

    for(beeIterator = bees.begin();
        beeIterator != bees.end();
        beeIterator++)
    {
        (*beeIterator)->updateBee(flowerLocations);

        if(!(*beeIterator)->getForgetFlowerThreshold())
        {
            zerolo ++;
        }

        if(!(*beeIterator)->getDanceFlowerThreshold())
        {
            zerohi ++;
        }

        if(doLogKnownFlowers)
        {
            const Position *knp = (*beeIterator)->getRealKnownFlowerPosition();

            if(knp != NULL)
                cout << knp->getXCoord() << " " << knp->getYCoord() << endl;
        }

        threshLow += (*beeIterator)->getForgetFlowerThreshold();
        threshHi += (*beeIterator)->getDanceFlowerThreshold();

        if(doLog && !(time%100))
            debugLogger->logBee(*beeIterator);
    }

/*    if(!(time % 500))
        cout << "T " << (float)threshLow/(bees.size()) << " " << (float)threshHi/(bees.size()) << " " << zerolo << " " << zerohi<< endl;
*/
    for(flowerIterator = flowers.begin();
        flowerIterator != flowers.end();)
    {
        (*flowerIterator)->updateFlower();

        deleteFlowerIterator = flowerIterator++;

        if(doLog && (*deleteFlowerIterator)->shouldMarkDead())
        {
            debugLogger->logFlower(*deleteFlowerIterator);
        }

        if((*deleteFlowerIterator)->canDelete())
        {
            //handle case where flowerIterator will go past flowers.end()
            //and top case won't catch it--this occurs when we delete the
            //last flower in flowers
            bool lastFlower = (flowerIterator == flowers.end());

            flowerLocations->deleteFlower(*deleteFlowerIterator);
            delete (*deleteFlowerIterator);
            flowers.erase(deleteFlowerIterator);

            if(lastFlower)
                break;
        }
    }

}

void World::addNBees(unsigned numBees)
{
    for(unsigned i = 0; i < numBees; ++i)
    {
        Position *newPosition = new Position(colony->getPosition()->getXCoord(),
                                             colony->getPosition()->getYCoord());
        Bee *newBee = new Bee(getNewGuid(), newPosition, FORAGER, colony, mySize);
        bees.push_back(newBee);
    }
}

void World::addBee(Bee *newBee)
{
    assert(colony);

    bees.push_back(newBee);
}

void World:: addFlower(Flower *newFlower)
{
    assert(colony);

    flowers.push_back(newFlower);
    flowerLocations->insertFlower(newFlower);

    if(doLog)
        debugLogger->logFlower(newFlower);
}

void World::addColony(Colony *newColony)
{
    colony = newColony;
}

void World::addNFlowers(unsigned numFlowers)
{
    for(unsigned i = 0; i < numFlowers; ++i)
    {
        Position *newPosition = new Position(rand() % mySize,
                                              rand() % mySize);

        Flower *newFlower = new Flower(getNewGuid(),
                                       newPosition,
                                       DEFAULT_FLOWER_LIFE,
                                       DEFAULT_FLOWER_QUANTITY,
                                       DEFAULT_FLOWER_QUALITY);

        if(doLog)
            debugLogger->logFlower(newFlower);

        flowers.push_back(newFlower);

        flowerLocations->insertFlower(newFlower);
    }
}

unsigned World::totalNectarCollected()
{
    return colony->getNectarCollected();
}

unsigned World::totalEnergySpent()
{
    return colony->getEnergySpent();
}

unsigned World::getDancesFollowed()
{
    return colony->getDancesFollowed();
}

unsigned World::getDancesOffered()
{
    return colony->getDancesOffered();
}

unsigned World::getNumScouts()
{
    return colony->getNumScouts();
}
